# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0


import json

# import traceback
from typing import Any, List, Union

from lxml import etree

from .. import utils
from ..analysers.properties_verifier import Verifier
from ..logger import Logger
from ..statics import constants


class JSONFileWebServiceListScraper(object):
    """
    A class to get all `tds2stac_webservice_analyser` in tag_config.json
    file when `tds2stac_mode_analyser` is `get` or `check`.
    """

    def __init__(self, json_file: str, logger_properties: dict = dict()):
        self.json_file = json_file
        self.logger_properties = (
            logger_properties if logger_properties is not None else {}
        )

    def load_and_process_json(self):
        json_webservice_set = set()
        try:
            with open(self.json_file) as file:
                loaded_json = json.load(file)

            for key, value in loaded_json.items():
                if isinstance(value, dict):
                    for key1, value1 in value.items():
                        if isinstance(value1, dict):
                            tds2stac_mode = value1.get(
                                "tds2stac_mode_analyser"
                            )
                            if tds2stac_mode in ["get", "check"]:
                                json_webservice_set.add(
                                    value1.get("tds2stac_webservice_analyser")
                                )
                                break

            del loaded_json
        except Exception as e:
            self.logger_properties["logger_level"] = "CRITICAL"
            self.logger_properties[
                "logger_msg"
            ] = f"Failed to process file {self.json_file}: {str(e)}"
            Logger(self.logger_properties)

        return list(json_webservice_set)


class WebServiceListScraper(object):
    """
    A class for getting the list of available web services
    of a TDS catalogs.

        Args:
            url (str): The catalog URL from TDS to provide its web services
            logger_properties (dict, optional):The dictionary of the logger properties.
            requests_properties (dict, optional): A dictionary that modify the requests to URLs.
    """

    url: str
    """
    url is the url of the TDS catalog
    """
    logger_properties: dict
    """
    The dictionary of the logger properties. You can look at keys in :class:`~tds2stac.logger.Logger` class.
    """
    requests_properties: dict
    """
    To obtain additional information on this topic, refer to the :class:`~tds2stac.TDS2STAC.requests_properties`. The default value is an empty dictionary.
    """

    def __init__(
        self,
        url: str,
        logger_properties: dict = dict(),
        requests_properties: dict = dict(),
    ):
        ###############################################
        verifier = Verifier()
        if logger_properties is not None and isinstance(
            logger_properties, dict
        ):
            verifier.logger_properties(logger_properties)
        if requests_properties is not None and isinstance(
            requests_properties, dict
        ):
            verifier.requests_properties(requests_properties)
        if logger_properties is not None and isinstance(
            logger_properties, dict
        ):
            self.logger_properties = logger_properties
        self.requests_properties = requests_properties
        self.logger_properties = logger_properties
        self.webserivces = []
        url_xml, id_xml, xml = utils.xml_processing(
            url, self.requests_properties
        )
        #################################
        # Getting the tree of the xml file
        #################################
        try:
            tree_data = etree.XML(xml)
        except Exception:
            self.logger_properties["logger_level"] = "CRITICAL"
            self.logger_properties[
                "logger_msg"
            ] = "The TDS Catalog URL does not provide any accessible services. Check it manually!"
            Logger(self.logger_properties)
            return
        else:
            try:
                ###############################################
                # Getting `dataset`, `catalog_url` and `catalog_id` of the data
                ###############################################
                dataset = tree_data.find("{%s}dataset" % constants.unidata)
                metadata = dataset.find("{%s}metadata" % constants.unidata)
                ########################################
                # Finding service tag in data tree element
                ########################################
                service_tag = dataset.find(
                    "{%s}serviceName" % constants.unidata
                )
                if service_tag is None and metadata is not None:
                    service_tag = metadata.find(
                        "{%s}serviceName" % constants.unidata
                    )
                if service_tag is None:
                    # Use services found in the file depends on the version of TDS 4 or 5
                    try:
                        services = tree_data.findall(
                            ".//{%s}service[@serviceType='Compound']"
                            % constants.unidata
                        )
                    except Exception:
                        services = tree_data.findall(
                            ".//{%s}service[@serviceType='compound']"
                            % constants.unidata
                        )
                else:
                    # Use specific named services
                    services = tree_data.findall(
                        ".//{%s}service[@name='%s']"
                        % (constants.unidata, service_tag.text)
                    )
                #############################################
                # Loop over all existing services in the data
                #############################################
                self.webserivces = [
                    s.get("name")
                    for i in services
                    for s in i.findall("{%s}service" % constants.unidata)
                    if (
                        i.get("serviceType") == "Compound"
                        or i.get("serviceType") == "compound"
                    )
                    # and s.get("name") in constants.supported_webservices
                ]
                self.webserivces = list(dict.fromkeys(self.webserivces))
            except Exception:
                self.logger_properties["logger_level"] = "CRITICAL"
                self.logger_properties[
                    "logger_msg"
                ] = "The TDS Catalog URL does not provide any accessible services!"
                Logger(self.logger_properties)
                # print(traceback.format_exc())
                return

    def aslist(self):
        return self.webserivces

    def __iter__(self):
        return iter(self.aslist())


class ConfigFileWebServicesScraper(object):
    """
    A class for getting the list of webservices that used
    in the `tag_config.json` file.
    Args:
        json_file (str): The path to the `tag_config.json` file
        logger_properties (dict): The dictionary of the logger properties.
    """

    json_file: str
    """
    The path to the `tag_config.json` file.To obtain further details on the creation of a
    `tag_config.json` file, refer: :ref:`tag-config`.
    """
    logger_properties: dict
    """
    The dictionary of the logger properties. You can look at keys in :class:`~tds2stac.logger.Logger` class.
    """

    def __init__(self, json_file: str, logger_properties: dict = dict()):
        self.json_file = json_file
        self.logger_properties = logger_properties
        self.values: list = []
        loaded_json = json.load(open(json_file))
        self.get_values(loaded_json, "tds2stac_webservice_analyser")

    def get_values(self, json_obj: Union[dict, list], key: str):
        """
        A function for getting the list of values of a
        specific key in a json file.
        """
        if isinstance(json_obj, dict):
            for k, v in json_obj.items():
                if k == key:
                    self.values.append(v)
                if isinstance(v, (dict, list)):
                    self.get_values(v, key)
        elif isinstance(json_obj, list):
            for item in json_obj:
                self.get_values(item, key)
        else:
            self.logger_properties["logger_level"] = "CRITICAL"
            self.logger_properties[
                "logger_msg"
            ] = f"The json file is not in a right format. Check the following file: {self.json_file}"
            Logger(self.logger_properties)

        return self.values

    def aslist(self):
        """
        A function for returning the list of webservices
        """
        self.values = list(dict.fromkeys(self.values))
        return self.values

    def __iter__(self):
        """
        A function for returning the iterator of webservices
        """
        return iter(self.aslist())


class WebServiceContentScraper(object):
    """
    The functionality of the existing class is dependent on the settings
    specified in the `tag_config.json` file in order to harvest targeted
    information from a selected web service. For comprehensive instructions
    on configuring the `tag_config.json` file, refer to the following link:
    :ref:`tag-config`.

        Args:
            root (etree._Element): The root of the XML-based web service
            json_file (str): The path to the `tag_config.json` file
            extensions_list (list): The list of extensions to be harvested
                from the web service (main keys in the `tag_config.json` file)
            harvesting_vars (dict, optional): The dictionary of harvesting variables
            logger_properties (dict, optional):The dictionary of the logger properties.
    """

    root: etree._Element
    """
    Etree root object of the XML-based web service
    """
    json_file: str
    """
    The path to the `tag_config.json` file
    """
    extensions_list: list
    """
    The list of extensions to be harvested from the web service.
    Main keys in the `tag_config.json` file. For example `item_datacube_extension`
    and so on.
    """
    harvesting_vars: Union[dict, None]
    """
    It's a dictionary that keys are variable names and values are the result of harvesting.
    """
    logger_properties: dict
    """
    The dictionary of the logger properties. You can look at keys in :class:`~tds2stac.logger.Logger` class.
    """

    def __init__(
        self,
        root: etree._Element,
        service_url: str,
        json_file: str,
        extensions_list: list,
        harvesting_vars: Union[dict, None] = None,
        logger_properties: dict = dict(),
    ):
        self.logger_properties = logger_properties
        tag_config_json = json.load(open(json_file))
        self.list_of_all_tags: List[Any] = []
        for extension in extensions_list:
            self.harvester(
                root, service_url, tag_config_json, extension, harvesting_vars
            )
        del tag_config_json

    def harvester(
        self, root, service_url, json_file, ext_name, harvesting_vars=None
    ):
        for k, v in json_file[ext_name].items():
            xpath_string_schema = ""
            namespaces_ = dict()
            action_type = []
            values_with_none_ = []
            list_of_all_tags_with_none = []
            if v is not None and isinstance(v, dict):
                # Main function starts from this point. because it finds the attrs of a tag
                if v.get("tds2stac_mode_analyser") == "str":
                    harvesting_vars[k] = v.get("tds2stac_manual_variable")
                    if harvesting_vars[k] is None:
                        self.logger_properties["logger_level"] = "WARNING"
                        self.logger_properties[
                            "logger_msg"
                        ] = f"The '{k}' is empty or in a wrong format. Check '{k}' key in your `tag_config.json` file."
                        Logger(self.logger_properties)
                elif v.get("tds2stac_mode_analyser") == "list":
                    Logger(self.logger_properties)
                    if v.get("tds2stac_manual_variable") is not None:
                        harvesting_vars[k] = (
                            v.get("tds2stac_manual_variable")
                            .strip("][")
                            .split(", ")
                        )
                    else:
                        self.logger_properties["logger_level"] = "WARNING"
                        self.logger_properties[
                            "logger_msg"
                        ] = f"The '{k}' is empty or in a wrong format. Check '{k}' key in your `tag_config.json` file."
                elif v.get("tds2stac_mode_analyser") == "get":
                    if v.get("tds2stac_webservice_analyser") is not None:
                        if (
                            f"/thredds/{v.get('tds2stac_webservice_analyser')}/"
                            in service_url
                        ):
                            # make a function to make a string to input to the xpath function
                            if v.get("tds2stac_reference_key") is not None:
                                self.logger_properties[
                                    "logger_level"
                                ] = "WARNING"
                                self.logger_properties[
                                    "logger_msg"
                                ] = f"key '{k}': Using `tds2stac_reference_key` is not recommended. It would be better better to use `tds2stac_webservice_analyser` instead."
                                Logger(self.logger_properties)
                                for k1, v1 in v.items():
                                    if (
                                        k
                                        not in constants.static_list_webservices
                                    ):
                                        # A condition specifically for ISO 19115-2 XML files

                                        if isinstance(v1, dict):
                                            # This condition is for finding which method should
                                            # be used to get the result data. For example, if
                                            # all attributed were field, it means we should use
                                            # `tag.text` method to get the result data. If one of
                                            # the attributes were None, it means we should use
                                            # tag.get(attr) method to get the result data.
                                            if v1 == list(v.values())[-1]:
                                                if (
                                                    list(v1.values()).count(
                                                        None
                                                    )
                                                    == 0
                                                ):
                                                    values_with_none_.append(
                                                        "text_of_tag"
                                                    )
                                                    action_type.append(
                                                        "text_of_tag"
                                                    )
                                                elif (
                                                    list(v1.values()).count(
                                                        None
                                                    )
                                                    == 1
                                                ):
                                                    values_with_none_.extend(
                                                        list(v1.values())
                                                    )
                                                    action_type.append(
                                                        list(v1.keys())[
                                                            list(
                                                                v1.values()
                                                            ).index(None)
                                                        ]
                                                    )
                                                elif (
                                                    list(v1.values()).count(
                                                        None
                                                    )
                                                    > 1
                                                ):
                                                    list_of_more_than_one_None = [
                                                        v
                                                        for i, v in enumerate(
                                                            list(v1.values())
                                                        )
                                                        if v is None
                                                    ]
                                                    values_with_none_.extend(
                                                        list_of_more_than_one_None
                                                    )
                                                    list_of_more_than_one_None = [
                                                        list(v1.keys())[i]
                                                        for i, v in enumerate(
                                                            list(v1.values())
                                                        )
                                                        if v is None
                                                    ]
                                                    action_type.extend(
                                                        list_of_more_than_one_None
                                                    )

                                k_ref = v["tds2stac_reference_key"]
                                v_ref = json_file[ext_name][
                                    v["tds2stac_reference_key"]
                                ]
                                for k1, v1 in v_ref.items():
                                    if (
                                        k_ref
                                        not in constants.static_list_webservices
                                    ):
                                        # A condition specifically for ISO 19115-2 XML files
                                        if ":" in k1:
                                            schema = k1.split(":")[0]
                                            localname = k1.split(":")[1]
                                        else:
                                            schema = v_ref[
                                                "tds2stac_webservice_analyser"
                                            ]
                                            localname = k1
                                        # In this condition we don't add any attribute to the xpath string
                                        if v1 is None:
                                            xpath_string_schema += "/%s:%s" % (
                                                schema,
                                                localname,
                                            )
                                            namespaces_[
                                                schema
                                            ] = constants.schemas_dicts[schema]
                                        elif isinstance(v1, dict):
                                            # This condition is for finding which method should
                                            # be used to get the result data. For example, if
                                            # all attributed were field, it means we should use
                                            # tag.text method to get the result data. If one of
                                            # the attributes were None, it means we should use
                                            # tag.get(attr) method to get the result data.

                                            attribute_str = ""  # defining an empty string for collecting all tag elements and attributes for xpath search in the following loop
                                            for k2, v2 in v1.items():
                                                if v2 is not None:
                                                    attribute_str += (
                                                        '[@%s="%s"]'
                                                        % (
                                                            k2,
                                                            v2,
                                                        )
                                                    )
                                                # this condition defined for times that we have need to get one of the attributes of a tag
                                                elif (
                                                    v2 is None
                                                    and list(
                                                        v1.values()
                                                    ).count(None)
                                                    == 1
                                                ):
                                                    attribute_str += (
                                                        "[@%s]" % (k2)
                                                    )
                                                # For this condition we have to trigger (because it's within a loop
                                                # and might issue more than one log) a warning to inform the user that he/she
                                                #  should choose one of the attributes of a tag to restrict the search to the desirable tag attribute
                                                #  otherwise it will return a nested list.of elements that each list in the nested list
                                                #  is content of getting attribute of a tag.
                                                # For example if we define like {'name': null, 'value': null} it will return a nested list
                                                # like  [[... ,... ], [... ,... ]].
                                                elif (
                                                    v2 is None
                                                    and list(
                                                        v1.values()
                                                    ).count(None)
                                                    > 1
                                                ):
                                                    attribute_str += (
                                                        "[@%s]" % (k2)
                                                    )
                                                    self.logger_properties[
                                                        "logger_level"
                                                    ] = "WARNING"
                                                    self.logger_properties[
                                                        "logger_msg"
                                                    ] = f"It contains more than one null in the nested dict of '{k}'. You have to choose one of them. Revise your `tag_config.json` file."
                                                    Logger(
                                                        self.logger_properties
                                                    )

                                            xpath_string_schema += (
                                                "/%s:%s%s"
                                                % (
                                                    schema,
                                                    localname,
                                                    attribute_str,
                                                )
                                            )
                                            namespaces_[
                                                schema
                                            ] = constants.schemas_dicts[schema]

                                if namespaces_ != {}:
                                    self.list_of_all_tags = root.xpath(
                                        xpath_string_schema,
                                        namespaces=namespaces_,
                                    )
                                    if self.list_of_all_tags != []:
                                        for a in self.list_of_all_tags:
                                            if a.xpath(".//*") != []:
                                                list_of_all_tags_with_none_temp = (
                                                    []
                                                )
                                                for xx in a.xpath(".//*"):
                                                    if (
                                                        values_with_none_[0]
                                                        in xx.attrib.values()
                                                    ):
                                                        list_of_all_tags_with_none_temp.append(
                                                            xx
                                                        )
                                                    else:
                                                        # self.logger_properties[
                                                        #     "logger_level"
                                                        # ] = "ERROR"
                                                        # self.logger_properties[
                                                        #     "logger_msg"
                                                        # ] = f"The `tds2stac_reference_key` of '{k}' key is not valid. Check the `tag_config.json` file."
                                                        # Logger(self.logger_properties)
                                                        continue
                                                list_of_all_tags_with_none.extend(
                                                    list_of_all_tags_with_none_temp
                                                )
                                                if (
                                                    list_of_all_tags_with_none_temp
                                                    == []
                                                ):
                                                    list_of_all_tags_with_none.append(
                                                        None
                                                    )
                                            else:
                                                list_of_all_tags_with_none.append(
                                                    None
                                                )

                                    if len(list_of_all_tags_with_none) == 0:
                                        # self.logger_properties[
                                        #     "logger_level"
                                        # ] = "WARNING"
                                        # self.logger_properties[
                                        #     "logger_msg"
                                        # ] = f"It couldn't find any tag for the considered tag element '{k}'. It means you should check the tag element name in the `tag_config.json` file."
                                        # Logger(self.logger_properties)
                                        continue
                                    elif (
                                        len(list_of_all_tags_with_none) == 1
                                        and action_type != []
                                    ):
                                        if action_type[0] == "text_of_tag":
                                            harvesting_vars[
                                                k
                                            ] = list_of_all_tags_with_none[
                                                0
                                            ].text
                                        else:
                                            harvesting_vars[
                                                k
                                            ] = list_of_all_tags_with_none[
                                                0
                                            ].get(
                                                action_type[0]
                                            )
                                    elif len(list_of_all_tags_with_none) > 1:
                                        if action_type[0] == "text_of_tag":
                                            harvesting_vars[k] = [
                                                a.text
                                                if a is not None
                                                else None
                                                for a in list_of_all_tags_with_none
                                            ]
                                        else:
                                            harvesting_vars[k] = [
                                                a.get(action_type[b])
                                                if a is not None
                                                else None
                                                for b in range(
                                                    len(action_type)
                                                )
                                                for a in list_of_all_tags_with_none
                                            ]
                            else:
                                for k1, v1 in v.items():
                                    if (
                                        k
                                        not in constants.static_list_webservices
                                    ):
                                        # A condition specifically for ISO 19115-2 XML files
                                        if ":" in k1:
                                            schema = k1.split(":")[0]
                                            localname = k1.split(":")[1]
                                        else:
                                            schema = v.get(
                                                "tds2stac_webservice_analyser"
                                            )
                                            localname = k1
                                        # In this condition we don't add any attribute to the xpath string
                                        if v1 is None:
                                            xpath_string_schema += "/%s:%s" % (
                                                schema,
                                                localname,
                                            )
                                            if schema is not None:
                                                namespaces_[
                                                    schema
                                                ] = constants.schemas_dicts[
                                                    schema
                                                ]
                                            else:
                                                self.logger_properties[
                                                    "logger_level"
                                                ] = "WARNING"
                                                self.logger_properties[
                                                    "logger_msg"
                                                ] = f"`tds2stac_webservice_analyser` of '{k}' key is None. Revise your `tag_config.json` file."
                                                Logger(self.logger_properties)
                                                return
                                        elif isinstance(v1, dict):
                                            # This condition is for finding which method should
                                            # be used to get the result data. For example, if
                                            # all attributed were field, it means we should use
                                            # tag.text method to get the result data. If one of
                                            # the attributes were None, it means we should use
                                            # tag.get(attr) method to get the result data.
                                            if v1 == list(v.values())[-1]:
                                                if (
                                                    list(v1.values()).count(
                                                        None
                                                    )
                                                    == 0
                                                ):
                                                    action_type.append(
                                                        "text_of_tag"
                                                    )
                                                elif (
                                                    list(v1.values()).count(
                                                        None
                                                    )
                                                    == 1
                                                ):
                                                    action_type.append(
                                                        list(v1.keys())[
                                                            list(
                                                                v1.values()
                                                            ).index(None)
                                                        ]
                                                    )
                                                elif (
                                                    list(v1.values()).count(
                                                        None
                                                    )
                                                    > 1
                                                ):
                                                    list_of_more_than_one_None = [
                                                        list(v1.keys())[i]
                                                        for i, v in enumerate(
                                                            list(v1.values())
                                                        )
                                                        if v is None
                                                    ]
                                                    action_type.extend(
                                                        list_of_more_than_one_None
                                                    )
                                            attribute_str = ""  # defining an empty string for collecting all tag elements and attributes for xpath search in the following loop
                                            for k2, v2 in v1.items():
                                                if v2 is not None:
                                                    attribute_str += (
                                                        '[@%s="%s"]'
                                                        % (
                                                            k2,
                                                            v2,
                                                        )
                                                    )
                                                # this condition defined for times that we have need to get one of the attributes of a tag
                                                elif (
                                                    v2 is None
                                                    and list(
                                                        v1.values()
                                                    ).count(None)
                                                    == 1
                                                ):
                                                    attribute_str += (
                                                        "[@%s]" % (k2)
                                                    )
                                                # For this condition we have to trigger (because it's within a loop
                                                # and might issue more than one log) a warning to inform the user that he/she
                                                #  should choose one of the attributes of a tag to restrict the search to the desirable tag attribute
                                                #  otherwise it will return a nested list.of elements that each list in the nested list
                                                #  is content of getting attribute of a tag.
                                                # For example if we define like {'name': null, 'value': null} it will return a nested list
                                                # like  [[... ,... ], [... ,... ]].
                                                elif (
                                                    v2 is None
                                                    and list(
                                                        v1.values()
                                                    ).count(None)
                                                    > 1
                                                ):
                                                    # we have to define a Logger for this warning

                                                    attribute_str += (
                                                        "[@%s]" % (k2)
                                                    )
                                                    self.logger_properties[
                                                        "logger_level"
                                                    ] = "WARNING"
                                                    self.logger_properties[
                                                        "logger_msg"
                                                    ] = f"It contains more than one null in the nested dict of '{k}' and you have to choose one of them. Revise your `tag_config.json` file."
                                                    Logger(
                                                        self.logger_properties
                                                    )
                                            xpath_string_schema += (
                                                "/%s:%s%s"
                                                % (
                                                    schema,
                                                    localname,
                                                    attribute_str,
                                                )
                                            )
                                            namespaces_[
                                                schema
                                            ] = constants.schemas_dicts[schema]

                                if namespaces_ != {}:
                                    self.list_of_all_tags = root.xpath(
                                        xpath_string_schema,
                                        namespaces=namespaces_,
                                    )
                                    if len(self.list_of_all_tags) == 0:
                                        # self.logger_properties[
                                        #     "logger_level"
                                        # ] = "ERROR"
                                        # self.logger_properties[
                                        #     "logger_msg"
                                        # ] = f"It couldn't find any tag for the considered tag element '{k}'. It means you should check the tag element name in the `tag_config.json` file."
                                        # Logger(self.logger_properties)
                                        continue
                                    elif (
                                        len(self.list_of_all_tags) == 1
                                        and action_type != []
                                    ):
                                        if action_type[0] == "text_of_tag":
                                            harvesting_vars[
                                                k
                                            ] = self.list_of_all_tags[0].text
                                        else:
                                            harvesting_vars[
                                                k
                                            ] = self.list_of_all_tags[0].get(
                                                action_type[0]
                                            )
                                    elif len(self.list_of_all_tags) > 1:
                                        if action_type[0] == "text_of_tag":
                                            harvesting_vars[k] = [
                                                a.text
                                                for a in self.list_of_all_tags
                                            ]
                                        else:
                                            harvesting_vars[k] = [
                                                a.get(action_type[b])
                                                for b in range(
                                                    len(action_type)
                                                )
                                                for a in self.list_of_all_tags
                                            ]
                elif v.get("tds2stac_mode_analyser") == "check":
                    if v.get("tds2stac_webservice_analyser") is not None:
                        if (
                            f"/thredds/{v.get('tds2stac_webservice_analyser')}/"
                            in service_url
                        ):
                            # make a function to make a string to input to the xpath function
                            for k1, v1 in v.items():
                                if k not in constants.static_list_webservices:
                                    if ":" in k1:
                                        schema = k1.split(":")[0]
                                        localname = k1.split(":")[1]
                                    else:
                                        schema = v[
                                            "tds2stac_webservice_analyser"
                                        ]
                                        localname = k1
                                    if v1 is None:
                                        xpath_string_schema += "/%s:%s" % (
                                            schema,
                                            localname,
                                        )
                                        namespaces_[
                                            schema
                                        ] = constants.schemas_dicts[schema]
                                    elif isinstance(v1, dict):
                                        attribute_str = ""
                                        for k2, v2 in v1.items():
                                            if v2 is not None:
                                                attribute_str += (
                                                    '[@%s="%s"]'
                                                    % (
                                                        k2,
                                                        v2,
                                                    )
                                                )
                                            elif (
                                                v2 is None
                                                and list(v1.values()).count(
                                                    None
                                                )
                                                == 1
                                            ):
                                                attribute_str += "[@%s]" % (k2)
                                            elif (
                                                v2 is None
                                                and v1.values().count(None) > 1
                                            ):
                                                attribute_str += "[@%s]" % (k2)
                                                self.logger_properties[
                                                    "logger_level"
                                                ] = "WARNING"
                                                self.logger_properties[
                                                    "logger_msg"
                                                ] = f"It contains more than one null in the nested dict of '{k}' and you have to choose one of them. Revise your `tag_config.json` file."
                                                Logger(self.logger_properties)

                                        xpath_string_schema += "/%s:%s%s" % (
                                            schema,
                                            localname,
                                            attribute_str,
                                        )
                                        namespaces_[
                                            schema
                                        ] = constants.schemas_dicts[schema]
                            if namespaces_ != {}:
                                self.list_of_all_tags = root.xpath(
                                    xpath_string_schema, namespaces=namespaces_
                                )
                                if len(self.list_of_all_tags) == 0:
                                    # self.logger_properties["logger_level"] = "WARNING"
                                    # self.logger_properties[
                                    #     "logger_msg"
                                    # ] = f"It couldn't find any tag for the considered tag element '{k}'. It means you should check the tag element name in the `tag_config.json` file."
                                    # Logger(self.logger_properties)
                                    continue

                                elif len(self.list_of_all_tags) >= 1:
                                    harvesting_vars[k] = v[
                                        "tds2stac_manual_variable"
                                    ]
                elif v.get("tds2stac_mode_analyser") is None:
                    self.logger_properties["logger_level"] = "WARNING"
                    self.logger_properties[
                        "logger_msg"
                    ] = f"The 'tds2stac_mode_analyser' of '{k}' is None. Revise your `tag_config.json` file."
                    Logger(self.logger_properties)



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0

from typing import Union

import requests
from lxml import etree

from .. import logger, utils
from ..statics import constants


class Recognizer(object):
    """
    A class for recognizing nine different
    and possible scenarios in management of
    TDS datasets. We will explain each scenario
    in the following.

        First scenario: Just `catalogRef` tags are located directly under the dataset element tag.
            tag `https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/catalog.xml` (nested)
        Second senarion: `CatalogRefs` are not under a dataset element tag and directly come below the `catalog`.
            `https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/sensor_catalog_ext.xml` (nested)
        Third scenario: One single `dataset` tag is located next to `CatalogRef` tags. All are under a `dataset` tag.
            `https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/chirps/catalog.xml` (nested)
        Fourth scenario: An empty datasets.
            `https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/bio_geo_chem_catalog_ext.xml` or `https://thredds.atmohub.kit.edu/thredds/catalog/snowfogs/catalog.xml`
        Fifth scenario: There is no `CatalogRef` tag and all are `dataset` tag. All of them are under a `dataset` tag.
            `https://thredds.imk-ifu.kit.edu/thredds/catalog/climate/raster/global/chelsa/v1.2/catalog.html`
        Sixth scenario: A single dataset
            `https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/catalog.xml?dataset=regclim/raster/global/era5/sfc/single/era5_sfc_20210101.nc`
        Seventh scenario: An aggregated dataset
            `https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/swabian_moses_2021.xml?dataset=swabian_moses_aggregation`
        Eighth scenario: A combination of `caralogRef` and `dataset` tags that is not under a `dataset` tag.It's similar to second scenario but with datasets
            `https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/transfer.xml`
        Ninth scenario: When we have a bunch of single `dataset` tags next to catalogref. It's similar to third scenario but with more datasets.
            `https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/hydrogfd/v3.0/catalog.xml` (nested)



    Args:
        main_catalog_url: TDS Catalog url to start harvesting
        nested_check: An option for checking nested datasets in TDS (True or False)
        auth: Authentication for TDS catalog e.g.('user', 'password')
        logger_properties: A dictionary for logger properties.
        requests_properties: A dictionary for requests properties.
    """

    main_catalog_url: str
    """
        TDS Catalog url to start harvesting (*)
    """
    nested_check: bool
    """
        An option for checking nested datasets in TDS (True or False) (optional)
    """
    logger_properties: dict
    """
        A dictionary for logger properties. For more information see :class:`~tds2stac.logger.Logger`
    """
    requests_properties: dict
    """
        To obtain additional information on this topic, refer to
        the :class:`~tds2stac.TDS2STAC.requests_properties`. The default value is
        an empty dictionary.
    """

    def __init__(
        self,
        main_catalog_url: str,
        nested_check: bool = False,
        logger_properties: dict = dict(),
        requests_properties: dict = dict(),
    ):
        if logger_properties is not None:
            self.logger_properties = logger_properties
        self.requests_properties = requests_properties
        self.all_dirs: list = []
        self.all_href: list = []
        self.all_dirs_extensions: list = []
        self.nested_check = nested_check
        self.nested_num: int = 0
        self.nested_num_temp: int = 0
        self.status: Union[str, None] = None

        # using 'xml_processing' we get the XML contents of catalog URL.
        xml_url_catalog, id_catalog, xml = utils.xml_processing(
            main_catalog_url, self.requests_properties
        )

        self.id_catalog = id_catalog
        self.xml_url_catalog = xml_url_catalog
        self.xml_url_catalog_temp = xml_url_catalog
        self.xml = xml
        self.recognition_function(self.xml_url_catalog, self.xml)
        self.logger_properties["logger_level"] = "INFO"
        self.logger_properties[
            "logger_msg"
        ] = f"{self.status, self.nested_num}"
        logger.Logger(self.logger_properties)

    def recognition_function(self, url: str, xml_content):
        """
        A function for recognizing number of scenarios
        in TDS
        """

        # Opening the Catalog url
        try:
            tree = etree.XML(xml_content)
            # return [e for e in tree]
        except BaseException:
            return (
                "The Catalog is not reachable. Check the Catalg URL in the TDS"
            )

        # we have nine different cases.
        # First, catalogRefs are under the dataset tag (https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/catalog.html) --> nested
        # and one without dataset tag directly with catalogRef(https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/sensor_catalog_ext.html) --> nested
        # the third one is the case that we have a single data next to catalogRefs (https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/chirps/catalog.html) --> nested
        # the fourth is empty datasets (https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/bio_geo_chem_catalog_ext.html or https://thredds.atmohub.kit.edu/thredds/catalog/snowfogs/catalog.html)
        # the fifth one is having all data under the dataset (https://thredds.imk-ifu.kit.edu/thredds/catalog/climate/raster/global/chelsa/v1.2/catalog.html)
        # the sixth one is a single data (https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/catalog.html?dataset=regclim/raster/global/era5/sfc/single/era5_sfc_20210101.nc)
        # the seventh is a aggregated data (https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/swabian_moses_2021.html?dataset=swabian_moses_aggregation)
        # the eighth case is combination of caralogRef and dataset that is not under a dataset tag like second case but with dataset (https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/transfer.html)
        # the ninth case is the case when we have a bunch of single data next to catalogref (https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/hydrogfd/v3.0/catalog.html) --> nested

        all_tags_list = [e for e in tree]

        if "catalogRef" not in str(all_tags_list) and "dataset" in str(
            all_tags_list
        ):
            # First and third to seventh cases are here
            # First case:
            if (
                tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata)
                == []
                and tree.findall(".//{%s}catalogRef" % constants.unidata) != []
            ):
                if self.nested_check is not True:
                    self.status = "First Scenario"
                    return self.status
                else:
                    self.nested_num = 0
                    self.nested_num_temp = 0
                    self.nested_checker(url)

                    for i in self.all_dirs:
                        if i.count("/") > self.nested_num:
                            self.nested_num = i.count("/")
                    self.nested_num = self.nested_num + 1
                    self.status = "First Scenario"
                    return self.status, self.nested_num
            # Third case:
            if (
                tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata)
                != []
                and len(
                    tree.findall(
                        ".//{%s}dataset[@urlPath]" % constants.unidata
                    )
                )
                == 1
                and tree.findall(".//{%s}catalogRef" % constants.unidata) != []
            ):
                if self.nested_check is not True:
                    self.status = "Third Scenario"
                    return self.status
                else:
                    self.nested_num = 0
                    self.nested_num_temp = 0
                    self.nested_checker(url)

                    for i in self.all_dirs:
                        if i.count("/") > self.nested_num:
                            self.nested_num = i.count("/")
                    self.nested_num = self.nested_num + 1
                    self.status = "Third Scenario"
                    return self.status, self.nested_num

            # Ninth case:
            if (
                tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata)
                != []
                and len(
                    tree.findall(
                        ".//{%s}dataset[@urlPath]" % constants.unidata
                    )
                )
                > 1
                and tree.findall(".//{%s}catalogRef" % constants.unidata) != []
            ):
                if self.nested_check is not True:
                    self.status = "Ninth Scenario"
                    return self.status
                else:
                    self.nested_num = 0
                    self.nested_num_temp = 0
                    self.nested_checker(url)

                    for i in self.all_dirs:
                        if i.count("/") > self.nested_num:
                            self.nested_num = i.count("/")
                    self.nested_num = self.nested_num + 1
                    self.status = "Ninth Scenario"
                    return self.status, self.nested_num
            # Fourth case:
            if (
                tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata)
                == []
                and tree.findall(".//{%s}catalogRef" % constants.unidata) == []
            ):
                self.status = "Fourth Scenario"
                return self.status
            if (
                tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata)
                != []
                and tree.findall(".//{%s}catalogRef" % constants.unidata) == []
            ):
                # Fifth to Seventh cases are here
                # THIS CONDITION SHOULD BE REFACTORED. I THINK BY REMOVING [@urlPath] FROM THE CONDITION IT WILL WORK
                if len(tree.findall(".//{%s}dataset" % constants.unidata)) > 1:
                    self.status = "Fifth Scenario"
                    return self.status
                else:
                    # Sixth and Seventh cases are here
                    dataset = tree.find("{%s}dataset" % constants.unidata)
                    metadata = dataset.find("{%s}metadata" % constants.unidata)

                    # Serviceså
                    service_tag = dataset.find(
                        "{%s}serviceName" % constants.unidata
                    )

                    if service_tag is None:
                        if metadata is not None:
                            service_tag = metadata.find(
                                "{%s}serviceName" % constants.unidata
                            )
                        else:
                            return "The dataset is not even Single or Aggregated dataset".__str__()

                    if service_tag is None:
                        # Use services found in the file. FMRC aggs do this.
                        services = tree.findall(
                            ".//{%s}service[@serviceType='Compound']"
                            % constants.unidata
                        )
                    else:
                        # Use specific named services    THIS PART SHOULD BE REFACTORED
                        services = tree.findall(
                            ".//{%s}service[@name='%s']"
                            % (constants.unidata, service_tag.text)
                        )

                    for i, service in enumerate(services):
                        # In TDS version 4 and 5 'Compound' is different
                        if (
                            service.get("serviceType") == "Compound"
                            or service.get("serviceType") == "compound"
                        ):
                            for s in service.findall(
                                "{%s}service" % constants.unidata
                            ):
                                if dataset.get("urlPath") is not None:
                                    service_url = utils.references_urls(
                                        url, s.get("base")
                                    ) + dataset.get("urlPath")
                                else:
                                    service_url = utils.references_urls(
                                        url, s.get("base")
                                    )
                                if s.get("name") == "http":
                                    a = requests.head(service_url)
                                    if "Content-Length" not in a.headers:
                                        self.status = "Seventh Scenario"
                                        return self.status
                                    else:
                                        self.status = "Sixth Scenario"
                                        return self.status

        # Second case:
        elif "catalogRef" in str(all_tags_list) and "dataset" not in str(
            all_tags_list
        ):
            if self.nested_check is not True:
                self.status = "Second Scenario"
                return self.status
            else:
                self.nested_num = 0
                self.nested_num_temp = 0
                self.nested_checker_exceptions(url)
                for i in self.all_href:
                    for j in self.all_href:
                        if i[0] in j[1]:
                            if j[0] not in self.all_dirs_extensions:
                                self.all_dirs_extensions.append(j[0])
                self.nested_num = len(self.all_dirs_extensions) + 1
                self.status = "Second Scenario"
                return self.status, self.nested_num
        # Eighth case:
        elif "catalogRef" in str(all_tags_list) and "dataset" in str(
            all_tags_list
        ):
            if self.nested_check is not True:
                self.status = "Eighth Scenario"
                return self.status
            else:
                self.nested_num = 0
                self.nested_num_temp = 0
                self.nested_checker_exceptions(url)
                for i in self.all_href:
                    for j in self.all_href:
                        if i[0] in j[1]:
                            if j[0] not in self.all_dirs_extensions:
                                self.all_dirs_extensions.append(j[0])

                self.nested_num = len(self.all_dirs_extensions) + 1
                self.status = "Eighth Scenario"
                return self.status, self.nested_num
        # Fourth case:
        elif "catalogRef" not in str(all_tags_list) and "dataset" not in str(
            all_tags_list
        ):
            if (
                tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata)
                == []
                and tree.findall(".//{%s}catalogRef" % constants.unidata) == []
            ):
                self.status = "Fourth Scenario"
                return self.status

    def nested_checker(self, url: str):
        """
        A function for returning the depth of
        nested datasets in TDS for scenarios 1, 3,
        ,and 9
        """
        xml_url_catalog, id_catalog, xml = utils.xml_processing(
            url, self.requests_properties
        )
        try:
            tree = etree.XML(xml)
        except BaseException:
            return (
                "The Catalog is not reachable. Check the Catalg URL in the TDS"
            )

        for child in tree.findall(".//{%s}catalogRef" % constants.unidata):
            if "catalogRef" in str(child):
                self.nested_checker(
                    utils.references_urls(
                        url, child.get("{%s}href" % constants.w3)
                    )
                )
                url_modified = url.replace("catalog.xml", "").replace(
                    self.xml_url_catalog.replace("catalog.xml", ""), ""
                )
                if url_modified not in self.all_dirs:
                    self.all_dirs.append(
                        url.replace("catalog.xml", "").replace(
                            self.xml_url_catalog.replace("catalog.xml", ""), ""
                        )
                    )

    def nested_checker_exceptions(self, url: str):
        """
        A function for returning the depth of
        nested datasets in TDS for scenarios 2 and 8
        """
        xml_url_catalog, id_catalog, xml = utils.xml_processing(
            url, self.requests_properties
        )

        try:
            tree = etree.XML(xml)
        except BaseException:
            return (
                "The Catalog is not reachable. Check the Catalg URL in the TDS"
            )

        for child in tree.findall(".//{%s}catalogRef" % constants.unidata):
            if "catalogRef" in str(child):
                self.nested_checker_exceptions(
                    utils.references_urls(
                        url, child.get("{%s}href" % constants.w3)
                    )
                )

                if (
                    url,
                    [
                        utils.references_urls(
                            url, c.get("{%s}href" % constants.w3)
                        )
                        for c in tree.findall(
                            ".//{%s}catalogRef" % constants.unidata
                        )
                    ],
                ) not in self.all_href:
                    self.all_href.append(
                        (
                            url,
                            [
                                utils.references_urls(
                                    url, c.get("{%s}href" % constants.w3)
                                )
                                for c in tree.findall(
                                    ".//{%s}catalogRef" % constants.unidata
                                )
                            ],
                        )
                    )

    def __str__(self):
        return f"{self.status, self.nested_num}"
